;;; ucs-normalize --- tests for international/ucs-normalize.el -*- lexical-binding: t -*-

;; Copyright (C) 2002-2016 Free Software Foundation, Inc.

;; This file is part of GNU Emacs.

;; GNU Emacs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;;; Code:

(eval-when-compile (require 'cl-lib))
(require 'ert)
(require 'ucs-normalize)

(defconst ucs-normalize-test-data-file
  (expand-file-name "test/data/NormalizationTest.txt" source-directory))

(defun ucs-normalize-tests--parse-column ()
  (let ((chars nil)
        (term nil))
    (while (and (not (equal term ";"))
                (looking-at "\\([[:xdigit:]]\\{4,6\\}\\)\\([; ]\\)"))
      (let ((code-point (match-string 1)))
        (setq term (match-string 2))
        (goto-char (match-end 0))
        (push (string-to-number code-point 16) chars)))
    (nreverse chars)))

(defmacro ucs-normalize-tests--normalize (norm str)
  "Like `ucs-normalize-string' but reuse current buffer for efficiency.
And NORM is one of the symbols `NFC', `NFD', `NFKC', `NFKD' for brevity."
  (let ((norm-alist '((NFC . ucs-normalize-NFC-region)
                      (NFD . ucs-normalize-NFD-region)
                      (NFKC . ucs-normalize-NFKC-region)
                      (NFKD . ucs-normalize-NFKD-region))))
    `(save-restriction
       (narrow-to-region (point) (point))
       (insert ,str)
       (funcall #',(cdr (assq norm norm-alist)) (point-min) (point-max))
       (delete-and-extract-region (point-min) (point-max)))))

(defun ucs-normalize-tests--invariants-hold-p (&rest columns)
  (cl-destructuring-bind (source nfc nfd nfkc nfkd)
      (mapcar (lambda (c) (apply #'string c)) columns)
    (and (equal nfc (ucs-normalize-tests--normalize NFC source))
         (equal nfc (ucs-normalize-tests--normalize NFC nfc))
         (equal nfc (ucs-normalize-tests--normalize NFC nfd))
         (equal nfkc (ucs-normalize-tests--normalize NFC nfkc))
         (equal nfkc (ucs-normalize-tests--normalize NFC nfkd))

         (equal nfd (ucs-normalize-tests--normalize NFD source))
         (equal nfd (ucs-normalize-tests--normalize NFD nfc))
         (equal nfd (ucs-normalize-tests--normalize NFD nfd))
         (equal nfkd (ucs-normalize-tests--normalize NFD nfkc))
         (equal nfkd (ucs-normalize-tests--normalize NFD nfkd))

         (equal nfkc (ucs-normalize-tests--normalize NFKC source))
         (equal nfkc (ucs-normalize-tests--normalize NFKC nfc))
         (equal nfkc (ucs-normalize-tests--normalize NFKC nfd))
         (equal nfkc (ucs-normalize-tests--normalize NFKC nfkc))
         (equal nfkc (ucs-normalize-tests--normalize NFKC nfkd))

         (equal nfkd (ucs-normalize-tests--normalize NFKD source))
         (equal nfkd (ucs-normalize-tests--normalize NFKD nfc))
         (equal nfkd (ucs-normalize-tests--normalize NFKD nfd))
         (equal nfkd (ucs-normalize-tests--normalize NFKD nfkc))
         (equal nfkd (ucs-normalize-tests--normalize NFKD nfkd)))))

(ert-deftest ucs-normalize-part0 ()
  (with-temp-buffer
    (insert-file-contents ucs-normalize-test-data-file)
    ;; Narrow to part.
    (let ((beg (progn (search-forward "@Part0") (forward-line) (point)))
          (end (progn (search-forward "@Part1") (match-beginning 0))))
      (delete-region end (point-max))   ; Delete from end first, so as
      (delete-region (point-min) beg)   ; not to ruin beg position.
      (goto-char (point-min)))
    (while (not (eobp))
      (unless (= (following-char) ?#)   ; Skip comments.
        (ucs-normalize-tests--invariants-hold-p
         (ucs-normalize-tests--parse-column)
         (ucs-normalize-tests--parse-column)
         (ucs-normalize-tests--parse-column)
         (ucs-normalize-tests--parse-column)
         (ucs-normalize-tests--parse-column)))
      (forward-line))))

(ert-deftest ucs-normalize-part1 ()
  (with-temp-buffer
    (insert-file-contents ucs-normalize-test-data-file)
    ;; Narrow to part.
    (narrow-to-region (progn (search-forward "@Part1") (forward-line) (point))
                      (progn (search-forward "@Part2") (match-beginning 0)))
    (goto-char (point-min))
    (let ((reporter (make-progress-reporter
                     "UCS Normalize Test Part1"
                     0 (count-lines (point-min) (point-max)) 0 nil 0.5))
          (line 0)
          (failed nil))
      (while (not (eobp))
        (unless (or (= (following-char) ?#)
                    (ucs-normalize-tests--invariants-hold-p
                     (ucs-normalize-tests--parse-column)
                     (ucs-normalize-tests--parse-column)
                     (ucs-normalize-tests--parse-column)
                     (ucs-normalize-tests--parse-column)
                     (ucs-normalize-tests--parse-column)))
          (push line failed))
        (cl-incf line)
        (progress-reporter-update reporter line)
        (forward-line))
      (progress-reporter-done reporter)
      (widen)
      (goto-char (point-min))
      (search-forward "@Part1")
      (forward-line)
      (let* ((start-line (count-lines (point-min) (point)))
             (failed-lines (mapcar (lambda (lino) (+ lino 1 start-line)) (nreverse failed))))
        (should-not failed-lines)))))

;;; ucs-normalize-tests.el ends here
